@@ -4,14 +4,18 @@ from click_default_group import DefaultGroup
 from dataclasses import asdict
 import io
 import json
+import re
 from llm import (
     Attachment,
+    AsyncConversation,
+    AsyncKeyModel,
     AsyncResponse,
     Collection,
     Conversation,
     Response,
     Template,
     UnknownModelError,
+    KeyModel,
     encode,
     get_async_model,
     get_default_model,
@@ -19,7 +23,6 @@ from llm import (
     get_embedding_models_with_aliases,
     get_embedding_model_aliases,
     get_embedding_model,
-    get_key,
     get_plugins,
     get_model,
     get_model_aliases,
@@ -30,6 +33,7 @@ from llm import (
     set_default_embedding_model,
     remove_alias,
 )
+from llm.models import _BaseConversation
 
 from .migrations import migrate
 from .plugins import pm, load_plugins
@@ -361,7 +365,7 @@ def prompt(
     if conversation_id or _continue:
         # Load the conversation - loads most recent if no ID provided
         try:
-            conversation = load_conversation(conversation_id)
+            conversation = load_conversation(conversation_id, async_=async_)
         except UnknownModelError as ex:
             raise click.ClickException(str(ex))
 
@@ -381,10 +385,6 @@ def prompt(
     except UnknownModelError as ex:
         raise click.ClickException(ex)
 
-    # Provide the API key, if one is needed and has been provided
-    if model.needs_key:
-        model.key = get_key(key, model.needs_key, model.key_env_var)
-
     if conversation:
         # To ensure it can see the key
         conversation.model = model
@@ -402,11 +402,16 @@ def prompt(
         except pydantic.ValidationError as ex:
             raise click.ClickException(render_errors(ex.errors()))
 
+    kwargs = {**validated_options}
+
     resolved_attachments = [*attachments, *attachment_types]
 
     should_stream = model.can_stream and not no_stream
     if not should_stream:
-        validated_options["stream"] = False
+        kwargs["stream"] = False
+
+    if isinstance(model, (KeyModel, AsyncKeyModel)):
+        kwargs["key"] = key
 
     prompt = read_prompt()
     response = None
@@ -424,7 +429,7 @@ def prompt(
                         prompt,
                         attachments=resolved_attachments,
                         system=system,
-                        **validated_options,
+                        **kwargs,
                     )
                     async for chunk in response:
                         print(chunk, end="")
@@ -435,7 +440,7 @@ def prompt(
                         prompt,
                         attachments=resolved_attachments,
                         system=system,
-                        **validated_options,
+                        **kwargs,
                     )
                     text = await response.text()
                     if extract or extract_last:
@@ -451,7 +456,7 @@ def prompt(
                 prompt,
                 attachments=resolved_attachments,
                 system=system,
-                **validated_options,
+                **kwargs,
             )
             if should_stream:
                 for chunk in response:
@@ -585,10 +590,6 @@ def chat(
     except KeyError:
         raise click.ClickException("'{}' is not a known model".format(model_id))
 
-    # Provide the API key, if one is needed and has been provided
-    if model.needs_key:
-        model.key = get_key(key, model.needs_key, model.key_env_var)
-
     if conversation is None:
         # Start a fresh conversation for this chat
         conversation = Conversation(model=model)
@@ -608,9 +609,15 @@ def chat(
         except pydantic.ValidationError as ex:
             raise click.ClickException(render_errors(ex.errors()))
 
+    kwargs = {}
+    kwargs.update(validated_options)
+
     should_stream = model.can_stream and not no_stream
     if not should_stream:
-        validated_options["stream"] = False
+        kwargs["stream"] = False
+
+    if key and isinstance(model, KeyModel):
+        kwargs["key"] = key
 
     click.echo("Chatting with {}".format(model.model_id))
     click.echo("Type 'exit' or 'quit' to exit")
@@ -641,7 +648,7 @@ def chat(
                 raise click.ClickException(str(ex))
         if prompt.strip() in ("exit", "quit"):
             break
-        response = conversation.prompt(prompt, system=system, **validated_options)
+        response = conversation.prompt(prompt, system=system, **kwargs)
         # System prompt only sent for the first message:
         system = None
         for chunk in response:
@@ -651,7 +658,9 @@ def chat(
         print("")
 
 
-def load_conversation(conversation_id: Optional[str]) -> Optional[Conversation]:
+def load_conversation(
+    conversation_id: Optional[str], async_=False
+) -> Optional[_BaseConversation]:
     db = sqlite_utils.Database(logs_db_path())
     migrate(db)
     if conversation_id is None:
@@ -668,11 +677,13 @@ def load_conversation(conversation_id: Optional[str]) -> Optional[Conversation]:
             "No conversation found with id={}".format(conversation_id)
         )
     # Inflate that conversation
-    conversation = Conversation.from_row(row)
+    conversation_class = AsyncConversation if async_ else Conversation
+    response_class = AsyncResponse if async_ else Response
+    conversation = conversation_class.from_row(row)
     for response in db["responses"].rows_where(
         "conversation_id = ?", [conversation_id]
     ):
-        conversation.responses.append(Response.from_row(db, response))
+        conversation.responses.append(response_class.from_row(db, response))
     return conversation
 
 
@@ -872,6 +883,9 @@ order by prompt_attachments."order"
 @click.option("-m", "--model", help="Filter by model or model alias")
 @click.option("-q", "--query", help="Search for logs matching this string")
 @click.option("-t", "--truncate", is_flag=True, help="Truncate long strings in output")
+@click.option(
+    "-s", "--short", is_flag=True, help="Shorter YAML output with truncated prompts"
+)
 @click.option("-u", "--usage", is_flag=True, help="Include token usage")
 @click.option("-r", "--response", is_flag=True, help="Just output the last response")
 @click.option("-x", "--extract", is_flag=True, help="Extract first fenced code block")
@@ -908,6 +922,7 @@ def logs_list(
     model,
     query,
     truncate,
+    short,
     usage,
     response,
     extract,
@@ -923,6 +938,16 @@ def logs_list(
     db = sqlite_utils.Database(path)
     migrate(db)
 
+    if short and (json_output or response):
+        invalid = " or ".join(
+            [
+                flag[0]
+                for flag in (("--json", json_output), ("--response", response))
+                if flag[1]
+            ]
+        )
+        raise click.ClickException("Cannot use --short and {} together".format(invalid))
+
     if response and not current_conversation and not conversation_id:
         current_conversation = True
 
@@ -1035,6 +1060,40 @@ def logs_list(
         current_system = None
         should_show_conversation = True
         for row in rows:
+            if short:
+                system = _truncate_string(row["system"], 120, end=True)
+                prompt = _truncate_string(row["prompt"], 120, end=True)
+                cid = row["conversation_id"]
+                attachments = attachments_by_id.get(row["id"])
+                obj = {
+                    "model": row["model"],
+                    "datetime": row["datetime_utc"].split(".")[0],
+                    "conversation": cid,
+                }
+                if system:
+                    obj["system"] = system
+                if prompt:
+                    obj["prompt"] = prompt
+                if attachments:
+                    items = []
+                    for attachment in attachments:
+                        details = {"type": attachment["type"]}
+                        if attachment.get("path"):
+                            details["path"] = attachment["path"]
+                        if attachment.get("url"):
+                            details["url"] = attachment["url"]
+                        items.append(details)
+                    obj["attachments"] = items
+                if usage and (row["input_tokens"] or row["output_tokens"]):
+                    usage_details = {
+                        "input": row["input_tokens"],
+                        "output": row["output_tokens"],
+                    }
+                    if row["token_details"]:
+                        usage_details["details"] = json.loads(row["token_details"])
+                    obj["usage"] = usage_details
+                click.echo(yaml.dump([obj], sort_keys=False).strip())
+                continue
             click.echo(
                 "# {}{}\n{}".format(
                     row["datetime_utc"].split(".")[0],
@@ -1115,15 +1174,22 @@ _type_lookup = {
     "--options", is_flag=True, help="Show options for each model, if available"
 )
 @click.option("async_", "--async", is_flag=True, help="List async models")
-@click.option("-q", "--query", help="Search for models matching this string")
+@click.option(
+    "-q",
+    "--query",
+    multiple=True,
+    help="Search for models matching these strings",
+)
 def models_list(options, async_, query):
     "List available models"
     models_that_have_shown_options = set()
     for model_with_aliases in get_models_with_aliases():
         if async_ and not model_with_aliases.async_model:
             continue
-        if query and not model_with_aliases.matches(query):
-            continue
+        if query:
+            # Only show models where every provided query string matches
+            if not all(model_with_aliases.matches(q) for q in query):
+                continue
         extra = ""
         if model_with_aliases.aliases:
             extra = " (aliases: {})".format(", ".join(model_with_aliases.aliases))
@@ -1255,17 +1321,51 @@ def aliases_list(json_):
 
 @aliases.command(name="set")
 @click.argument("alias")
-@click.argument("model_id")
-def aliases_set(alias, model_id):
+@click.argument("model_id", required=False)
+@click.option(
+    "-q",
+    "--query",
+    multiple=True,
+    help="Set alias for model matching these strings",
+)
+def aliases_set(alias, model_id, query):
     """
     Set an alias for a model
 
     Example usage:
 
     \b
-        $ llm aliases set turbo gpt-3.5-turbo
+        llm aliases set mini gpt-4o-mini
+
+    Alternatively you can omit the model ID and specify one or more -q options.
+    The first model matching all of those query strings will be used.
+
+    \b
+        llm aliases set mini -q 4o -q mini
     """
-    set_alias(alias, model_id)
+    if not model_id:
+        if not query:
+            raise click.ClickException(
+                "You must provide a model_id or at least one -q option"
+            )
+        # Search for the first model matching all query strings
+        found = None
+        for model_with_aliases in get_models_with_aliases():
+            if all(model_with_aliases.matches(q) for q in query):
+                found = model_with_aliases
+                break
+        if not found:
+            raise click.ClickException(
+                "No model found matching query: " + ", ".join(query)
+            )
+        model_id = found.model.model_id
+        set_alias(alias, model_id)
+        click.echo(
+            f"Alias '{alias}' set to model '{model_id}'",
+            err=True,
+        )
+    else:
+        set_alias(alias, model_id)
 
 
 @aliases.command(name="remove")
@@ -1536,6 +1636,10 @@ def embed(
 )
 @click.option("--prefix", help="Prefix to add to the IDs", default="")
 @click.option("-m", "--model", help="Embedding model to use")
+@click.option(
+    "--prepend",
+    help="Prepend this string to all content before embedding",
+)
 @click.option("--store", is_flag=True, help="Store the text itself in the database")
 @click.option(
     "-d",
@@ -1555,6 +1659,7 @@ def embed_multi(
     batch_size,
     prefix,
     model,
+    prepend,
     store,
     database,
 ):
@@ -1677,11 +1782,15 @@ def embed_multi(
         def tuples() -> Iterable[Tuple[str, Union[bytes, str]]]:
             for row in rows:
                 values = list(row.values())
-                id = prefix + str(values[0])
+                id: str = prefix + str(values[0])
+                content: Optional[Union[bytes, str]] = None
                 if binary:
-                    yield id, cast(bytes, values[1])
+                    content = cast(bytes, values[1])
                 else:
-                    yield id, " ".join(v or "" for v in values[1:])
+                    content = " ".join(v or "" for v in values[1:])
+                if prepend and isinstance(content, str):
+                    content = prepend + content
+                yield id, content or ""
 
         embed_kwargs = {"store": store}
         if batch_size:
@@ -1773,11 +1882,20 @@ def embed_models():
 
 
 @embed_models.command(name="list")
-def embed_models_list():
+@click.option(
+    "-q",
+    "--query",
+    multiple=True,
+    help="Search for embedding models matching these strings",
+)
+def embed_models_list(query):
     "List available embedding models"
     output = []
     for model_with_aliases in get_embedding_models_with_aliases():
-        s = str(model_with_aliases.model.model_id)
+        if query:
+            if not all(model_with_aliases.matches(q) for q in query):
+                continue
+        s = str(model_with_aliases.model)
         if model_with_aliases.aliases:
             s += " (aliases: {})".format(", ".join(model_with_aliases.aliases))
         output.append(s)
@@ -1897,10 +2015,17 @@ def template_dir():
     return path
 
 
-def _truncate_string(s, max_length=100):
-    if len(s) > max_length:
+def _truncate_string(s, max_length=100, end=False):
+    if not s:
+        return s
+    if end:
+        s = re.sub(r"\s+", " ", s)
+        if len(s) <= max_length:
+            return s
         return s[: max_length - 3] + "..."
-    return s
+    if len(s) <= max_length:
+        return s
+    return s[: max_length - 3] + "..."
 
 
 def logs_db_path():
